#pragma once

#include "eopcc/cloud.hpp"
#include "eopcc/io.hpp"

#include <iostream>

namespace eopcc {

enum class AttributeType {
  Unknown,
  X,
  Y,
  Z,
  Intensity,
  Red,
  Green,
  Blue,
  Timestamp
};

enum class DataType {
  unknown,
  uint8,
  int8,
  uint16,
  int16,
  uint32,
  int32,
  uint64,
  int64,
  float32,
  float64
};


inline AttributeType StringToAttributeType(const std::string& attr) {
  if (attr.compare("x") == 0) {
    return AttributeType::X;
  } else if (attr.compare("y") == 0) {
    return AttributeType::Y;
  } else if (attr.compare("z") == 0) {
    return AttributeType::Z;
  } else if (attr.compare("intensity") == 0) {
    return AttributeType::Intensity;
  } else if (attr.compare("red") == 0) {
    return AttributeType::Red;
  } else if (attr.compare("blue") == 0) {
    return AttributeType::Blue;
  } else if (attr.compare("green") == 0) {
    return AttributeType::Green;
  } else if (attr.compare("timestamp") == 0) {
    return AttributeType::Timestamp;
  }
  return AttributeType::Unknown;
}

inline DataType StringToDataType(const std::string& dtype) {
  if (dtype.compare("uchar") == 0) {
    return DataType::uint8;
  } else if (dtype.compare("char") == 0) {
    return DataType::int8;
  } else if (dtype.compare("ushort") == 0) {
    return DataType::uint16;
  } else if (dtype.compare("short") == 0) {
    return DataType::int16;
  } else if (dtype.compare("int") == 0) {
    return DataType::int32;
  } else if (dtype.compare("uint") == 0) {
    return DataType::uint32;
  } else if (dtype.compare("float") == 0) {
    return DataType::float32;
  } else if (dtype.compare("double") == 0) {
    return DataType::float64;
  }
  return DataType::unknown;
}

template<typename T> void AttributeIntoCloud(PointCloud& pc, const int index, const AttributeType attr, T val) {
  switch(attr) {
      case AttributeType::X:
        pc.loc[index].x = val;
        break;
      case AttributeType::Y:
        pc.loc[index].y = val;
        break;
      case AttributeType::Z:
        pc.loc[index].z = val;
        break;
      case AttributeType::Intensity:
        pc.intensity[index] = val;
        break;
      case AttributeType::Red:
        pc.color[index].red = val / 255.0;
        break;
      case AttributeType::Green:
        pc.color[index].green = val / 255.0;
        break;
      case AttributeType::Blue:
        pc.color[index].blue = val / 255.0;
        break;
      case AttributeType::Timestamp:
        pc.gps_time[index] = val;
        break;
  }
}

template<typename T> T ReadFromAscii(std::ifstream& in_file) {
   T val;
   in_file >> val;
   return val;
}

template<> uint8_t ReadFromAscii<uint8_t>(std::ifstream& in_file);

template<typename T> T ReadFromBinary(std::ifstream& in_file) {
  union {
    char bytes[sizeof(T)];
    T val;
  } buffer;
  buffer.val = 0;
  in_file.read((char*) buffer.bytes, sizeof(T));
  return buffer.val;
}

template<typename T> void ReadIntoCloud(std::ifstream& in_file, bool is_binary, PointCloud& pc,
                                        const int64_t index, const AttributeType attr) {
  T val;
  if (is_binary) {
    val = ReadFromBinary<T>(in_file);
  } else {
    val = ReadFromAscii<T>(in_file);
  }

  AttributeIntoCloud<T>(pc, index, attr, val);
}


class PLYReader : public PointCloudReader {
public:
  PLYReader(const std::string& filename);
  virtual PointCloud ReadSubCloud(int sub_cloud_size);
  virtual size_t NumRemainingPoints() { return num_points_remaining_; }
  virtual PointCloud GetSeedCloud() {
    return seed_cloud_;
  }
private:
  void ParseHeader();
  void SetupSeedCloud();
  void ReadPoint(PointCloud& pc, int index);

  bool is_binary_;
  std::ifstream ply_file_;
  size_t num_points_remaining_;
  int num_properties_;
  PointCloud seed_cloud_;

  std::vector<AttributeType> property_attr_;
  std::vector<DataType> property_types_;
};

class PLYWriter : public PointCloudWriter {
public:
  // The seed cloud determines which attributes to include in the header
  PLYWriter(const std::string& filename, const PointCloud& seed_cloud, bool is_binary);
  virtual ~PLYWriter();
  virtual void WriteSubCloud(const PointCloud& sub_cloud);

private:
  std::ofstream out_file_;
  bool is_binary_;
  bool has_time_;
  bool has_color_;
  size_t points_written_;
  size_t header_count_pos_;
};


// These methods can be used to force load PLY files with the wrong extension, but otherwise
// use the methods from cloud_io.hpp
void LoadFromPLY(const std::string& ply_file_name, PointCloud& point_cloud);
void WriteToPLY(const std::string& ply_file_name, const PointCloud& point_cloud, bool use_binary=false);


} // end namespace eopcc
